/* XmGoBang: (c) 1995-2008 by Andreas S. Krebs */

#include "xmgobang.h"

#define COMPUTER_MOVETIME	10	/* leave a little time to system */

/* some local prototypes */

static void initBoard();
static void computerOnlyGameTimer(XtPointer, XtIntervalId *);

static void makeComputerMove();
static void calculateNextMove(short *, short *, short);
static Boolean checkFor5(short, short, short *, short *, short *, short *);

static long getFieldValue(short, short, short);
static long getAreaValue(DirInfo *);
static void getDirectionInfo(short, short, short, short, DirInfo *); 


/* some local variables */

static GoBoard XGoBoard; 

/************************************************************************
 *									*
 * Callback procedure for the `undo`-button:				*
 *   this procedure undoes the last done move				*
 *									*
 ************************************************************************/

void undoCB(Widget widget , XtPointer client_data, XtPointer call_data)
{
  if(XGoBoard.MoveNumber>0)
  {
    XGoBoard.MoveNumber--;    
    undoChipAt(XGoBoard.MovesDone[XGoBoard.MoveNumber*2],
	       XGoBoard.MovesDone[XGoBoard.MoveNumber*2+1]);

    if(XGoGame.GameStatus!=GAME_PAUSED)
      setGameStatus(GAME_PAUSED);
    
    setPlayersTurn(XGoGame.PlayersTurn ^ 1);     

    if(XGoBoard.MoveNumber==0)
      switchOffWidget(UNDO_WIDGET);
    if(XGoBoard.MoveNumber+1 == XGoBoard.LastMoveDone)
      switchOnWidget(REDO_WIDGET);
  }
}


/************************************************************************
 *									*
 * Callback procedure for the `redo`-button:    			*
 *   this procedure redoes the last undone move 			*
 *									*
 ************************************************************************/

void redoCB(Widget widget , XtPointer client_data, XtPointer call_data)
{
  if(XGoBoard.LastMoveDone > XGoBoard.MoveNumber)
  {
    drawChipAt(XGoBoard.MovesDone[XGoBoard.MoveNumber*2],
	       XGoBoard.MovesDone[XGoBoard.MoveNumber*2+1], XGoGame.PlayersTurn);
    
    XGoBoard.MoveNumber++;    
    setPlayersTurn(XGoGame.PlayersTurn ^ 1);
    
    if(XGoBoard.MoveNumber >= XGoBoard.LastMoveDone)
      switchOffWidget(REDO_WIDGET);
    if(XGoBoard.MoveNumber == 1)
      switchOnWidget(UNDO_WIDGET);
  }
}


/************************************************************************
 *									*
 * Callback procedure for the `hint`-button:				*
 *   this procedure gives a human player a hint for the next move	*
 *									*
 ************************************************************************/

void hintCB(Widget widget , XtPointer client_data, XtPointer call_data)
{
  if(XGoGame.GameStatus!=GAME_PAUSED)
    setGameStatus(GAME_PAUSED);
    
  if(XGoBoard.MoveNumber < XGoBoard.NumberOfFields)
  {
    short x,y;
    
    calculateNextMove(&x, &y, XGoGame.PlayersTurn);
    drawChipAt(x, y, PLAYER_HINT);
  }
  else     
    switchOffWidget(HINT_WIDGET);
}


/************************************************************************
 *									*
 * Callback procedure for the `pause`-button: 				*
 *   this procedure changes the GameStatus depending on the current	*
 *   GameStatus								*
 *									*
 ************************************************************************/

void pauseCB(Widget widget, XtPointer client_data, XtPointer call_data)
{
  switch(XGoGame.GameStatus)
  {
    case GAME_PAUSED:
      if(XGoBoard.MoveNumber != XGoBoard.LastMoveDone)
      {
        switchOffWidget(REDO_WIDGET);
        XGoBoard.LastMoveDone=XGoBoard.MoveNumber;
      }
      
    case GAME_NEW:
      setGameStatus(GAME_RUNNING);
      startPlayerTimer(XGoGame.PlayersTurn);

      if(XGoGame.Player[0].Type==PLAYER_COMPUTER && 
	 XGoGame.Player[1].Type==PLAYER_COMPUTER)
      {	
	XtAppAddTimeOut(context, COMPUTER_MOVETIME,
        		(XtTimerCallbackProc)computerOnlyGameTimer, NULL);		    
      }
      else if(XGoGame.Player[XGoGame.PlayersTurn].Type==PLAYER_COMPUTER)
	makeComputerMove(XGoGame.PlayersTurn);
      break;
    case GAME_RUNNING:
      setGameStatus(GAME_PAUSED);
      break;
    case GAME_FINISHED:
      newGame();
      break;
  }  
}


/************************************************************************
 *									*
 * initialize a new game						*
 *									*
 ************************************************************************/

void newGame()
{
  setGameStatus(GAME_NEW);
  initBoard();

  switchOffWidget(UNDO_WIDGET);
  switchOffWidget(REDO_WIDGET);
  switchOffWidget(HINT_WIDGET);    

  drawBoard();

  XGoGame.Player[0].Seconds=0;
  setPlayerTimer(0);
  XGoGame.Player[1].Seconds=0;
  setPlayerTimer(1);

  setPlayersTurn(XGoGame.StartPlayer);
}


/************************************************************************
 *									*
 * alloc memory for a the game board and the moves stack		*
 *									*
 ************************************************************************/

void allocBoard()
{
  XGoBoard.BoardField=(char *)malloc(sizeof(char)*XGoGame.Columns*XGoGame.Rows);  
  
  XGoBoard.MovesDone=(char *)malloc(sizeof(char)*XGoGame.Columns*XGoGame.Rows*2);    
}


/************************************************************************
 *									*
 * free memory for a the game board and the moves stack			*
 *									*
 ************************************************************************/

void freeBoard()
{
  free(XGoBoard.BoardField);
  free(XGoBoard.MovesDone);
}


/************************************************************************
 *									*
 * initialize board by clearing all fields and resetting moves stack	*
 *									*
 ************************************************************************/

static void initBoard()
{
  short i,j;
  char *field;
 
  field=XGoBoard.BoardField;
  
  for(i=0;i<XGoGame.Rows;i++)
    for(j=0;j<XGoGame.Columns;j++)
      *field++=FIELD_EMPTY;		/* at beg. all field are empty */

  XGoBoard.MoveNumber=0;
  XGoBoard.LastMoveDone=0;
  XGoBoard.NumberOfFields=XGoGame.Rows*XGoGame.Columns;
}


/************************************************************************
 *									*
 * start a timer to make computer moves					*
 *									*
 ************************************************************************/

static void computerOnlyGameTimer(XtPointer client_data, XtIntervalId *idAdr)
{
  if(XGoGame.GameStatus==GAME_RUNNING)
  {
    makeComputerMove();
    
    XtAppAddTimeOut(context, COMPUTER_MOVETIME, 
		    (XtTimerCallbackProc)computerOnlyGameTimer, NULL);		    
  }
}

/************************************************************************
 *									*
 * place a "button" on the specified field		 		*
 *									*
 ************************************************************************/

void setField(short row, short col, char value)
{
  char *field;
  
  field=XGoBoard.BoardField;
  field+=(col * XGoGame.Columns + row);
  
  *field=value;
}


/************************************************************************
 *									*
 * return type of the "button" on the specified field		 	*
 *									*
 ************************************************************************/

char getField(short row, short col)
{
  if(row<0 || row>=XGoGame.Rows || col<0 || col>=XGoGame.Columns)
    return FIELD_BORDER;
  else
  {
    char *field;

    field=XGoBoard.BoardField;
    field+=(col * XGoGame.Columns + row);
  
    return *field;
  }
}


/************************************************************************
 *									*
 * make a move and change the PlayersTurn 			 	*
 *									*
 ************************************************************************/

void makeMove(short i, short j)
{
  Boolean won;
  short s_x, s_y, d_x, d_y;

  XGoBoard.MovesDone[XGoBoard.MoveNumber*2]=(unsigned char)i;
  XGoBoard.MovesDone[XGoBoard.MoveNumber*2+1]=(unsigned char)j;
  
  drawChipAt(i, j, XGoGame.PlayersTurn);
  
  if(XGoBoard.MoveNumber==0)
    switchOnWidget(UNDO_WIDGET);
  
  XGoBoard.MoveNumber++;
  XGoBoard.LastMoveDone=XGoBoard.MoveNumber;

  if(XGoGame.GameStatus!=GAME_RUNNING)
    setGameStatus(GAME_RUNNING);

  if(XGoGame.Player[XGoGame.PlayersTurn^1].Type==PLAYER_COMPUTER && 
     XGoGame.Player[XGoGame.PlayersTurn].Type==PLAYER_COMPUTER)
  {
    XGoGame.PlayersTurn^=1;
    /* don't display new turn, cause computer only game is to fast */
  }
  else
    setPlayersTurn(XGoGame.PlayersTurn^1); 
    
  won=checkFor5(i, j, &s_x, &s_y, &d_x, &d_y);
  if(won==True || XGoBoard.MoveNumber >= XGoBoard.NumberOfFields)
  {
    setGameStatus(GAME_FINISHED);
    switchOffWidget(UNDO_WIDGET);
    setCursor(GAME_FINISHED);		/* no more moves possible */

    displayGameOver(won, XGoGame.PlayersTurn^1, s_x, s_y, d_x, d_y);
    return;
  }

    
  if(XGoGame.Player[XGoGame.PlayersTurn^1].Type==PLAYER_HUMAN && 
     XGoGame.Player[XGoGame.PlayersTurn].Type==PLAYER_COMPUTER)
  {
    makeComputerMove(XGoGame.PlayersTurn);
  }
}


/************************************************************************
 *									*
 * let computer make a move  					 	*
 *									*
 ************************************************************************/

static void makeComputerMove()
{
  short x, y;

  if(XGoBoard.MoveNumber < XGoBoard.NumberOfFields)
  {
    XGoGame.Player[XGoGame.PlayersTurn].Seconds++;
    /* computer has a one second handicap per move ;-) */
    
    if(XGoGame.Player[XGoGame.PlayersTurn^1].Type==PLAYER_HUMAN)
      setPlayerTimer(XGoGame.PlayersTurn);
    /* don't display timer in computer only game, cause it's to fast */
    
    calculateNextMove(&x, &y, XGoGame.PlayersTurn); 
    makeMove(x, y);
  }
  else
  {
    setGameStatus(GAME_FINISHED);
      
    switchOffWidget(UNDO_WIDGET);
  }
}


/************************************************************************
 *									*
 * calculate the next move for player playerNr 			 	*
 *									*
 ************************************************************************/

static void calculateNextMove(short *x, short *y, short playerNr)
{
  short 	i, j;
  long	 	value=0, tmpval;
  int 		rand_num=1;
  
  srand(time(NULL));
  
  for(j=0;j<XGoGame.Columns;j++)
  {
    for(i=0;i<XGoGame.Rows;i++)
    {
      if(getField(i,j)==FIELD_EMPTY)
      {
        tmpval=getFieldValue(i, j, playerNr);
	
	if(tmpval > value)
	{
	  value=tmpval;
	  *x=i;
	  *y=j;
	  
	  rand_num=1;
	}
	else if(tmpval==value)
	{
	  rand_num++;
	  if(rand() < RAND_MAX/rand_num)
	  {
	    *x=i;
	    *y=j;
	  }
	}
      }
    }      
  }
}


/************************************************************************
 *									*
 * get value of this field, the field with the highest value is the 	*
 * best move to do							*
 *									*
 ************************************************************************/

static long getFieldValue(short x, short y, short playerNr)
{
  DirInfo me_dir, other_dir;
  long value=0, v;
  short k;
  
  for(k=0;k<4;k++)
  {
    getDirectionInfo(k, x, y, playerNr, &me_dir);
    getDirectionInfo(k, x, y, playerNr^1, &other_dir);  
  
    if(me_dir.SameColor>=4)
    {  
      value+=500000;	/* if I have a change to win, I use it ;-) */
      continue;
    }  
    else if(other_dir.SameColor>=4)
    {
      value+=100000;	/* avoid losing where ever possible */
      continue;
    }

    v=getAreaValue(&me_dir);	 
    if(XGoGame.PlayMode==PLAY_OFFENSIVE)
    {
      v*=11;
      v/=10;
    }
    value+=v;
    
    v=getAreaValue(&other_dir);
    if(XGoGame.PlayMode==PLAY_DEFENSIVE)
    {
      v*=11;
      v/=10;
    }
    value+=v;
  }
  
  return(value);
}


/************************************************************************
 *									*
 * get value of the area in the specified direction		 	*
 *									*
 ************************************************************************/

static long getAreaValue(DirInfo *di)
{
  long value=0;
  short h;  

  if(di->OpenEnds==0 || di->SizeOfArea<5)
    return 0;
  
  value=1;
  h=di->SameColor;
  while(h--)
    value*=10;

  h=di->SameInArea;
  while(h--)
    value*=3;
  
  if(di->OpenEnds>1)
    value*=11;
 
  if(di->SizeOfArea>5)
    value*=2;
  
  return value;
}


/************************************************************************
 *									*
 * calculate the next move for player playerNr 			 	*
 *									*
 ************************************************************************/

static short dirAdd[][2]=
{
  { 1, 0 },
  { 1, -1 },
  { 0, 1 },
  { 1, 1 }
};

static void getDirectionInfo(short dir, short x, short y, short playerNr,
			     DirInfo *di)
{
  short 	x_add, y_add, x_pos, y_pos;
  short 	i, dirs_done;
  
  x_add=dirAdd[dir][0];
  y_add=dirAdd[dir][1];
  
  di->SameColor=0;
  di->SameInArea=0;
  di->SizeOfArea=1;
  di->OpenEnds=0;
  dirs_done=0; 
  
  while(dirs_done<2)
  {  
    i=0;
    x_pos=x+x_add;
    y_pos=y+y_add;
    while(getField(x_pos, y_pos)==(playerNr+1))
    {
      i++;
      di->SizeOfArea++;
      di->SameColor++;
    
      x_pos+=x_add;
      y_pos+=y_add;
    }
  
    if(getField(x_pos, y_pos)==FIELD_EMPTY)
    {
      i++;
      di->OpenEnds++;
      di->SizeOfArea++;

      x_pos+=x_add;
      y_pos+=y_add;

      while(5-i>0 && getField(x_pos, y_pos)==(playerNr+1))
      {
	i++;
	di->SizeOfArea++;
	di->SameInArea++;

	x_pos+=x_add;
	y_pos+=y_add;
      }
      
      while(5-i>0 && (getField(x_pos, y_pos)==FIELD_EMPTY))
      {
 	i++;
	di->SizeOfArea++;
	
	x_pos+=x_add;
	y_pos+=y_add;
      }
    }
    
    x_add=-x_add;
    y_add=-y_add;
    dirs_done++;
  }  
}


/************************************************************************
 *									*
 * check if player has won by making move <x>, <y> 		 	*
 *									*
 ************************************************************************/

static Boolean checkFor5(short x, short y, 
			 short *x1, short *y1, short *x2, short *y2)
{
  DirInfo di;
  short k, player;
 
  if(getField(x,y)==FIELD_PLAYER1)
    player=0;
  else if(getField(x,y)==FIELD_PLAYER2)
    player=1;
  else
    return False;
  
  for(k=0;k<4;k++)
  {
    getDirectionInfo(k, x, y, player, &di);
			     
    if(di.SameColor>=4)
    { 
      *x1=x;
      *y1=y;
      while(getField(*x1+dirAdd[k][0], *y1+dirAdd[k][1])==player+1)
      {
 	*x1+=dirAdd[k][0];
	*y1+=dirAdd[k][1];
      }
      
      *x2=x;
      *y2=y;
      while(getField(*x2-dirAdd[k][0], *y2-dirAdd[k][1])==player+1)
      {
 	*x2-=dirAdd[k][0];
	*y2-=dirAdd[k][1];
      }
            
      return True;
    }  
  }
  
  return False;
}
